/*
 * Simulator of microcontrollers (dump.cc)
 *
 * Copyright (C) 1999,99 Drotos Daniel, Talker Bt.
 * 
 * To contact author send email to drdani@mazsola.iit.uni-miskolc.hu
 *
 */

/* This file is part of microcontroller simulator: ucsim.

UCSIM is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

UCSIM is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with UCSIM; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA. */
/*@1@*/

#include "ddconfig.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "i_string.h"

#include "simcl.h"

#include "uc51cl.h" //FIXME
#include "regs51.h" //FIXME
#include "globals.h"
#include "cmdutil.h"


/*
 * Dump address and instruction at specified addr
 */
//FIXME
void
dump_disass(class t_uc51 *uc, uint addr, class cl_sim *sim)
{
  char *dis;
  class cl_brk *b;
  int i;

  b  = uc->fbrk_at(addr);
  dis= uc->disass(addr, NULL);
  if (b)
    sim->cmd->printf("%c", (b->perm == brkFIX)?'F':'D');
  else
    sim->cmd->printf(" ");
  sim->cmd->printf("%c %06x %02x",
		   uc->inst_at(addr)?' ':'*',
		   addr, uc->get_mem(MEM_ROM, addr));
  for (i= 1; i < uc->inst_length(uc->get_mem(MEM_ROM, addr)); i++)
    sim->cmd->printf(" %02x", uc->get_mem(MEM_ROM, addr+i));
  while (i < 3)
    {
      sim->cmd->printf("   ");
      i++;
    }
  sim->cmd->printf(" %s\n", dis);
  free(dis);
}


/*
 * Dump memory
 */

void
dump_memory(cl_mem *mem,
	    uint *start, uint stop, int bpl, FILE *f,
	    class cl_sim *sim)
{
  int i, idx;

  while ((*start <= stop) &&
	 (*start < mem->size))
    {
      sim->cmd->printf("%06x ", *start);
      for (i= 0; (i < bpl) &&
	     (*start+i < mem->size) &&
	     (*start+i <= stop);
	   i++)
	{
	  char format[10];
	  sprintf(format, "%%0%dx ", mem->width/4);
	  fprintf(f, format/*"%02x "*/, mem->get(*start+i));
	}
      while (i < bpl)
	{
	  fprintf(f, "   ");
	  i++;
	}
      for (i= 0; (i < bpl) &&
	     (*start+i < mem->size) &&
	     (*start+i <= stop);
	   i++)
	fprintf(f, "%c", isprint(mem->get(*start+i))?mem->get(*start+i):'.');
      fprintf(f, "\n");
      (*start)+= bpl;
    }
}


/*
 * DISASSEMBLE [start [offset [lines]]]
 */
//FIXME
static disass_last_stop= 0;

bool
cmd_disass(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  char *s;
  int  start, offset= -1, dir;
  int  lines= 20;
  uint realstart;

  if (!uc->there_is_inst())
    return(FALSE);
  if ((s= strtok(NULL, delimiters)) != NULL)
    start= strtol(s, NULL, 0);
  else
    start= disass_last_stop;
  if ((s= strtok(NULL, delimiters)) != NULL)
    offset= strtol(s, NULL, 0);
  if ((s= strtok(NULL, delimiters)) != NULL)
    lines= strtol(s, NULL, 0);

  realstart= start;
  while (!uc->inst_at(realstart))
    realstart= (realstart+1) & (EROM_SIZE-1);
  if (offset)
    {
      dir= (offset < 0)?-1:+1;
      while (offset)
	{
	  realstart= (realstart+dir) & (EROM_SIZE-1);
	  while (!uc->inst_at(realstart))
	    realstart= (realstart+dir) & (EROM_SIZE-1);
	  offset+= -dir;
	}
    }
  
  while (lines)
    {
      dump_disass(uc, realstart, sim);
      realstart= (realstart+1) & (EROM_SIZE-1);
      while (!uc->inst_at(realstart))
	realstart= (realstart+1) & (EROM_SIZE-1);
      lines--;
    }

  disass_last_stop= realstart;
  return(FALSE);
}


/*
 * DUMP CODE [start [stop]]
 */
//FIXME
static int dc_last_stop= 0;

bool
cmd_dump_code(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  char *s;
  int  start= 0, stop= -1;
  int  lines= 20;
  
  if ((s= strtok(NULL, delimiters)) != NULL)
    start= (int)strtol(s, NULL, 0);
  else
    start= dc_last_stop;
  if ((s= strtok(NULL, delimiters)) != NULL)
    {
      stop = (int)strtol(s, NULL, 0);
      lines= -1;
    }
  else
    stop = -1;
  while (((stop >= 0) && (start <= stop)) ||
	 (lines >= 0))
    {
      dump_disass(uc, start, sim);
      start= (start+
	      uc->inst_length(uc->get_mem(MEM_ROM, start)))&(EROM_SIZE-1);
      lines--;
    }
  dc_last_stop= start;
  return(FALSE);
}


/*
 * DUMP CODE HEX [start [stop]]
 */

static int dch_last_stop= 0;

bool
cmd_dump_code_hex(char *cmd, class cl_uc *uc, class cl_sim *sim)
{
  char *s;
  int  start= 0, stop= -1;
  
  if ((s= strtok(NULL, delimiters)) != NULL)
    start= (int)strtol(s, NULL, 0);
  else
    start= dch_last_stop;
  if ((s= strtok(NULL, delimiters)) != NULL)
    stop = (int)strtol(s, NULL, 0);
  else
    stop = start + (10*16) - 1;
  dump_memory(uc->mem(MEM_ROM), (uint *)&start, stop, 16,
	      sim->cmd_out(), sim);
  dch_last_stop= start;
  return(FALSE);
}


/*
 * DUMP IRAM [start [stop]]
 */

static int di_last_stop= 0;

bool
cmd_dump_iram(char *cmd, class cl_uc *uc, class cl_sim *sim)
{
  char *s;
  int  start= 0, stop= -1;
  
  if ((s= strtok(NULL, delimiters)) != NULL)
    start= (int)strtol(s, NULL, 0);
  else
    start= di_last_stop;
  if ((s= strtok(NULL, delimiters)) != NULL)
    stop = (int)strtol(s, NULL, 0);
  else
    stop = start + (10*8) - 1;
  dump_memory(uc->mem(MEM_IRAM), (uint *)&start, stop, 8, sim->cmd_out(), sim);
  di_last_stop= start;
  return(FALSE);
}


/*
 * DUMP XRAM [start [stop]]
 */
//FIXME
static int dx_last_stop= 0;

bool
cmd_dump_xram(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  char *s;
  int  start= 0, stop= -1;
  
  if ((s= strtok(NULL, delimiters)) != NULL)
    start= (int)strtol(s, NULL, 0);
  else
    start= dx_last_stop;
  if ((s= strtok(NULL, delimiters)) != NULL)
    stop= (int)strtol(s, NULL, 0);
  else
    stop= start + (10*16) - 1;
  uc->eram2xram();
  dump_memory(uc->mem(MEM_XRAM), (uint *)&start, stop, 16, sim->cmd_out(), sim);
  dx_last_stop= start;
  return(FALSE);
}


/*
 * DUMP REGISTERS
 */
//FIXME
bool
cmd_dump_regs(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  uint  start;
  uchar data;

  start= uc->get_mem(MEM_SFR, PSW) & 0x18;
  dump_memory(uc->mem(MEM_IRAM), &start, start+7, 8, sim->cmd_out(), sim);
  start= uc->get_mem(MEM_SFR, PSW) & 0x18;
  data= uc->get_mem(MEM_IRAM, uc->get_mem(MEM_IRAM, start));
  sim->cmd->printf("%06x %02x %c",
		   uc->get_mem(MEM_IRAM, start), data, isprint(data)?data:'.');

  sim->cmd->printf("  ACC= 0x%02x %3d %c  B= 0x%02x",
		   uc->get_mem(MEM_SFR, ACC), uc->get_mem(MEM_SFR, ACC),
		   isprint(uc->get_mem(MEM_SFR, ACC))?
		   (uc->get_mem(MEM_SFR, ACC)):'.',
		   uc->get_mem(MEM_SFR, B)); 
  uc->eram2xram();
  data= uc->get_mem(MEM_XRAM,
		    uc->get_mem(MEM_SFR, DPH)*256+uc->get_mem(MEM_SFR, DPL));
  sim->cmd->printf("   DPTR= 0x%02x%02x @DPTR= 0x%02x %3d %c\n",
		   uc->get_mem(MEM_SFR, DPH), uc->get_mem(MEM_SFR, DPL),
		   data, data, isprint(data)?data:'.');

  data= uc->get_mem(MEM_IRAM, uc->get_mem(MEM_IRAM, start+1));
  sim->cmd->printf("%06x %02x %c", uc->get_mem(MEM_IRAM, start+1), data,
	  isprint(data)?data:'.');
  data= uc->get_mem(MEM_SFR, PSW);
  sim->cmd->printf("  PSW= 0x%02x CY=%c AC=%c OV=%c P=%c\n", data,
		   (data&bmCY)?'1':'0', (data&bmAC)?'1':'0',
		   (data&bmOV)?'1':'0', (data&bmP)?'1':'0');

  dump_disass(uc, uc->PC, sim);
  return(FALSE);
}


/*
 * DUMP PORTS
 */
//FIXME
bool
cmd_dump_port(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  uchar data;

  data= uc->get_mem(MEM_SFR, P0);
  sim->cmd->printf("P0    ");
  print_bin(data, 8, sim->cmd_out());
  sim->cmd->printf(" 0x%02x %3d %c", data, data, isprint(data)?data:'.');

  data= uc->get_mem(MEM_SFR, P1);
  sim->cmd->printf("    P1    ");
  print_bin(data, 8, sim->cmd_out());
  sim->cmd->printf(" 0x%02x %3d %c\n", data, data, isprint(data)?data:'.');

  data= uc->port_pins[0];
  sim->cmd->printf("Pin0  ");
  print_bin(data, 8, sim->cmd_out());
  sim->cmd->printf(" 0x%02x %3d %c", data, data, isprint(data)?data:'.');

  data= uc->port_pins[1];
  sim->cmd->printf("    Pin1  ");
  print_bin(data, 8, sim->cmd_out());
  sim->cmd->printf(" 0x%02x %3d %c\n", data, data, isprint(data)?data:'.');

  data= uc->port_pins[0] & uc->get_mem(MEM_SFR, P0);
  sim->cmd->printf("Port0 ");
  print_bin(data, 8, sim->cmd_out());
  sim->cmd->printf(" 0x%02x %3d %c", data, data, isprint(data)?data:'.');

  data= uc->port_pins[1] & uc->get_mem(MEM_SFR, P1);
  sim->cmd->printf("    Port1 ");
  print_bin(data, 8, sim->cmd_out());
  sim->cmd->printf(" 0x%02x %3d %c\n", data, data, isprint(data)?data:'.');

  sim->cmd->printf("\n");

  data= uc->get_mem(MEM_SFR, P2);
  sim->cmd->printf("P2    ");
  print_bin(data, 8, sim->cmd_out());
  sim->cmd->printf(" 0x%02x %3d %c", data, data, isprint(data)?data:'.');

  data= uc->get_mem(MEM_SFR, P3);
  sim->cmd->printf("    P3    ");
  print_bin(data, 8, sim->cmd_out());
  sim->cmd->printf(" 0x%02x %3d %c\n", data, data, isprint(data)?data:'.');

  data= uc->port_pins[2];
  sim->cmd->printf("Pin2  ");
  print_bin(data, 8, sim->cmd_out());
  sim->cmd->printf(" 0x%02x %3d %c", data, data, isprint(data)?data:'.');

  data= uc->port_pins[3];
  sim->cmd->printf("    Pin3  ");
  print_bin(data, 8, sim->cmd_out());
  sim->cmd->printf(" 0x%02x %3d %c\n", data, data, isprint(data)?data:'.');

  data= uc->port_pins[2] & uc->get_mem(MEM_SFR, P2);
  sim->cmd->printf("Port2 ");
  print_bin(data, 8, sim->cmd_out());
  sim->cmd->printf(" 0x%02x %3d %c", data, data, isprint(data)?data:'.');

  data= uc->port_pins[3] & uc->get_mem(MEM_SFR, P3);
  sim->cmd->printf("    Port3 ");
  print_bin(data, 8, sim->cmd_out());
  sim->cmd->printf(" 0x%02x %3d %c\n", data, data, isprint(data)?data:'.');

  return(FALSE);
}


/*
 * DUMP SFR [addr...]
 */
//FIXME
bool
cmd_dump_sfr(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  char *s;
  uint start= 0x80;
  uchar data;
  struct name_entry *ne;

  if ((s= strtok(NULL, delimiters)) != NULL)
    while (s != NULL)
      {
	if ((ne= get_name_entry(uc->sfr_tbl(), s, uc)) != NULL)
	  {
	    data= uc->get_mem(MEM_SFR, ne->addr);
	    sim->cmd->printf("%6s %02x %3d %c\n", ne->name, data, data,
			     isprint(data)?data:'.');
	  }
	else
	  {
	    start= strtol(s, NULL, 0);
	    data = uc->get_mem(MEM_SFR, start);
	    if (start >= SFR_START)
	      sim->cmd->printf("%06x %02x %3d %c\n", start, data, data,
			       isprint(data)?data:'.');
	  }
	s= strtok(NULL, delimiters);
      }
  else
    // dump all
    dump_memory(uc->mem(MEM_SFR), &start, 255, 16, sim->cmd_out(), sim);
  return(FALSE);
}


/*
 * DUMP BIT addr...
 */
//FIXME
bool
cmd_dump_bit(char *cmd, class t_uc51 *uc, class cl_sim *sim)
{
  char *s, *sym;
  uchar *cell, addr;
  uchar bitaddr, bitmask;

  if ((s= strtok(NULL, delimiters)) == NULL)
    {
      sim->cmd->printf("Address has not given.\n");
      return(FALSE);
    }
  while (s)
    {
      if (!interpret_bitname(s, uc, &cell, &addr, &bitaddr, &bitmask, &sym))
	{
	  sim->cmd->printf("Bad address %s\n", s);
	  return(FALSE);
	}
      sim->cmd->printf("%06x %6s %c\n", addr, sym, (*cell & bitmask)?'1':'0');
      free(sym);
      s= strtok(NULL, delimiters);
    }
  return(FALSE);
}


/* End of dump.cc */
