/*
 * Simulator of microcontrollers (uc89c51r.cc)
 *
 * Copyright (C) 1999,99 Drotos Daniel, Talker Bt.
 * 
 * To contact author send email to drdani@mazsola.iit.uni-miskolc.hu
 *
 */

/* This file is part of microcontroller simulator: ucsim.

UCSIM is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

UCSIM is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with UCSIM; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA. */
/*@1@*/

#include "ddconfig.h"

#include <stdio.h>

// local
#include "uc89c51rcl.h"
#include "regs51.h"


t_uc89c51r::t_uc89c51r(int Itype, int Itech, class cl_sim *asim):
  t_uc51r(Itype, Itech, asim)
{
  it_sources->add_at(4, new cl_it_src(bmEC, CCON, bmCCF4, 0x0033, false,
				      "PCA module #4"));
  it_sources->add_at(4, new cl_it_src(bmEC, CCON, bmCCF3, 0x0033, false,
				      "PCA module #3"));
  it_sources->add_at(4, new cl_it_src(bmEC, CCON, bmCCF2, 0x0033, false,
				      "PCA module #2"));
  it_sources->add_at(4, new cl_it_src(bmEC, CCON, bmCCF1, 0x0033, false,
				      "PCA module #1"));
  it_sources->add_at(4, new cl_it_src(bmEC, CCON, bmCCF0, 0x0033, false,
				      "PCA module #0"));
  it_sources->add_at(4, new cl_it_src(bmEC, CCON, bmCF, 0x0033, false,
				      "PCA counter"));
}


void
t_uc89c51r::reset(void)
{
  t_uc51r::reset();
  MEM(MEM_SFR)[CCAPM0]|= bmECOM;
  MEM(MEM_SFR)[CCAPM1]|= bmECOM;
  MEM(MEM_SFR)[CCAPM2]|= bmECOM;
  MEM(MEM_SFR)[CCAPM3]|= bmECOM;
  MEM(MEM_SFR)[CCAPM4]|= bmECOM;
  t0_overflows= 0;
  dpl0= dph0= dpl1= dph1= 0;
  MEM(MEM_SFR)[IPH]= 0;
}

void
t_uc89c51r::proc_write(uchar *addr)
{
  t_uc51r::proc_write(addr);
  if (addr == &(MEM(MEM_SFR)[CCAP0L]))
    MEM(MEM_SFR)[CCAPM0]&= ~bmECOM;
  if (addr == &(MEM(MEM_SFR)[CCAP0H]))
    MEM(MEM_SFR)[CCAPM0]|= bmECOM;
  if (addr == &(MEM(MEM_SFR)[CCAP1L]))
    MEM(MEM_SFR)[CCAPM1]&= ~bmECOM;
  if (addr == &(MEM(MEM_SFR)[CCAP1H]))
    MEM(MEM_SFR)[CCAPM1]|= bmECOM;
  if (addr == &(MEM(MEM_SFR)[CCAP2L]))
    MEM(MEM_SFR)[CCAPM2]&= ~bmECOM;
  if (addr == &(MEM(MEM_SFR)[CCAP2H]))
    MEM(MEM_SFR)[CCAPM2]|= bmECOM;
  if (addr == &(MEM(MEM_SFR)[CCAP3L]))
    MEM(MEM_SFR)[CCAPM3]&= ~bmECOM;
  if (addr == &(MEM(MEM_SFR)[CCAP3H]))
    MEM(MEM_SFR)[CCAPM3]|= bmECOM;
  if (addr == &(MEM(MEM_SFR)[CCAP4L]))
    MEM(MEM_SFR)[CCAPM4]&= ~bmECOM;
  if (addr == &(MEM(MEM_SFR)[CCAP4H]))
    MEM(MEM_SFR)[CCAPM4]|= bmECOM;

  if (addr == &(MEM(MEM_SFR)[AUXR]))
    MEM(MEM_SFR)[AUXR]&= ~(0x04);
}

uchar
t_uc89c51r::read(uchar *addr)
{
  return(t_uc51r::read(addr));
}

int
t_uc89c51r::it_priority(uchar ie_mask)
{
  uchar l, h;

  l= MEM(MEM_SFR)[IP] & ie_mask;
  h= MEM(MEM_SFR)[IPH] & ie_mask;
  if (!h && !l)
    return(0);
  if (!h && l)
    return(1);
  if (h && !l)
    return(2);
  if (h && l)
    return(3);
}

void
t_uc89c51r::pre_inst(void)
{
  if (MEM(MEM_SFR)[AUXR1] & bmDPS)
    {
      MEM(MEM_SFR)[DPL]= dpl1;
      MEM(MEM_SFR)[DPH]= dph1;
    }
  else
    {
      MEM(MEM_SFR)[DPL]= dpl0;
      MEM(MEM_SFR)[DPH]= dph0;
    }
}

void
t_uc89c51r::post_inst(void)
{
  if (MEM(MEM_SFR)[AUXR1] & bmDPS)
    {
      dpl1= MEM(MEM_SFR)[DPL];
      dph1= MEM(MEM_SFR)[DPH];
    }
  else
    {
      dpl0= MEM(MEM_SFR)[DPL];
      dph0= MEM(MEM_SFR)[DPH];
    }
}


/*
 * Simulating timers
 *
 * Calling inherited method to simulate timer #0, #1, #2 and then 
 * simulating Programmable Counter Array
 */

int
t_uc89c51r::do_timers(int cycles)
{
  int res;

  if ((res= t_uc51r::do_timers(cycles)) != resGO)
    return(res);
  return(do_pca(cycles));
}

int
t_uc89c51r::t0_overflow(void)
{
  uchar cmod= get_mem(MEM_SFR, CMOD) & (bmCPS0|bmCPS1);

  if (cmod == bmCPS1)
    t0_overflows++;
}


/*
 * Simulating Programmable Counter Array
 */

int
t_uc89c51r::do_pca(int cycles)
{
  uchar ccon;
  int ret= resGO;

  if (!((ccon= get_mem(MEM_SFR, CCON)) & bmCR))
    return(resGO);
  if (state == stIDLE &&
      (ccon & bmCIDL))
    return(resGO);

  switch (get_mem(MEM_SFR, CMOD) & (bmCPS1|bmCPS0))
    {
    case 0:
      ret= do_pca_counter(cycles);
      break;
    case bmCPS0:
      ret= do_pca_counter(cycles*3);
      break;
    case bmCPS1:
      ret= do_pca_counter(t0_overflows);
      t0_overflows= 0;
      break;
    case (bmCPS0|bmCPS1):
      if ((prev_p1 & bmECI) != 0 &
	  (MEM(MEM_SFR)[P1] & bmECI) == 0)
	do_pca_counter(1);
      break;
    }
  return(ret);
}

int
t_uc89c51r::do_pca_counter(int cycles)
{
  while (cycles--)
    {
      if (++(MEM(MEM_SFR)[CL]) == 0)
	{
	  if (++(MEM(MEM_SFR)[CH]) == 0)
	    {
	      /* CH,CL overflow */
	      MEM(MEM_SFR)[CCON]|= bmCF;
	      do_pca_module(0);
	      do_pca_module(1);
	      do_pca_module(2);
	      do_pca_module(3);
	      do_pca_module(4);
	    }
	}
    }
  return(resGO);
}

int
t_uc89c51r::do_pca_module(int nr)
{
  uchar CCAPM[5]= {0xda, 0xdb, 0xdc, 0xdd, 0xde};
  uchar CCAPL[5]= {0xea, 0xeb, 0xec, 0xed, 0xee};
  uchar CCAPH[5]= {0xfa, 0xfb, 0xfc, 0xfd, 0xfe};
  uchar bmCEX[5]= {bmCEX0, bmCEX1, bmCEX2, bmCEX3, bmCEX4};
  uchar bmCCF[5]= {bmCCF0, bmCCF1, bmCCF2, bmCCF3, bmCCF4};
  uchar ccapm= MEM(MEM_SFR)[CCAPM[nr]];
  
  if (
      ((ccapm & bmCAPP) &&
       (prev_p1 & bmCEX[nr]) == 0 &&
       (MEM(MEM_SFR)[P1] & bmCEX[nr]) != 0)
      ||
      ((ccapm & bmCAPN) &&
       (prev_p1 & bmCEX[nr]) != 0 &&
       (MEM(MEM_SFR)[P1] & bmCEX[nr]) == 0)
      )
    {
      /* Capture */
      MEM(MEM_SFR)[CCAPL[nr]]= MEM(MEM_SFR)[CL];
      MEM(MEM_SFR)[CCAPH[nr]]= MEM(MEM_SFR)[CH];
      MEM(MEM_SFR)[CCON]|= bmCCF[nr];
    }

  if (ccapm & bmECOM)
    {
      /* Comparator enabled */
      if (MEM(MEM_SFR)[CL] == MEM(MEM_SFR)[CCAPL[nr]] &&
	  MEM(MEM_SFR)[CH] == MEM(MEM_SFR)[CCAPH[nr]])
	{
	  /* Match */
	  if (nr == 4 &&
	      (MEM(MEM_SFR)[CMOD] & bmWDTE))
	    {
	      reset();
	    }
	  MEM(MEM_SFR)[CCON]|= bmCCF[nr];
	  if (ccapm & bmTOG)
	    {
	      /* Toggle */
	      MEM(MEM_SFR)[P1]^= bmCEX[nr];
	    }
	}
      if (ccapm & bmPWM)
	{
	  /* PWM */
	  if (MEM(MEM_SFR)[CL] == 0)
	    MEM(MEM_SFR)[CCAPL[nr]]= MEM(MEM_SFR)[CCAPH[nr]];
	  if (MEM(MEM_SFR)[CL] < MEM(MEM_SFR)[CCAPL[nr]])
	    MEM(MEM_SFR)[P1]&= ~(bmCEX[nr]);
	  else
	    MEM(MEM_SFR)[P1]|= bmCEX[nr];
	}
    }

  return(resGO);
}


/* End of uc89c51r.cc */
